import shlex
import json
import logging
import subprocess
from . import v3
from . import v4
from .base import BaseApi
from .errors import InvalidApiCall


logger = logging.getLogger('civicrmapi')


class BaseConsoleApi(BaseApi):
    """
    Base class for CiviCRM Console API implementations. Subclasses must define the
    :attr:`~base.BaseApi.VERSION` attribute and implement the :meth:`._get_command` method.

    :raises NotImplemented: when :attr:`~base.BaseApi.VERSION` is not defined
    :raises NotImplemented: when :meth:`._get_command` is not implemented
    """

    # FIXME: Instead of using cwd, cv --cwd=/path/to/civicrm could be used directly
    # as cv parameter. There should be a hint within the docs.
    def __init__(self, cv='cv', cwd=None, context=None):
        super().__init__()
        self.cv = shlex.split(cv)
        self.cwd = ['--cwd', shlex.quote(f'{cwd}')] if cwd else list()
        self.context = shlex.split(context) if context else None

    def _get_command(self, entity, action, params):
        raise NotImplemented

    def _perform_api_call(self, entity, action, params):
        command = self._get_command(entity, action, params)

        # To run the command within a given context we tokenize the original
        # command and give it as positional argument to the context.
        if self.context:
            command = f'{" ".join(self.context)} {shlex.quote(command)}'

        # Log the command.
        logger.info(f'Run command: {command}')

        # Run the command.
        try:
            reply = subprocess.run(command, shell=True, capture_output=True, check=True, text=True)

        except subprocess.CalledProcessError as exc:
            # Unfortunately we have no chance to identify invalid api calls by
            # just the return code. So we need some further diggin.

            # If we have json formatted stdout we probably have an invalid api
            # v3 call.
            try:
                data = json.loads(exc.stdout)
            except json.JSONDecodeError:
                pass
            else:
                raise InvalidApiCall(data)

            # An invalid api v4 call should be contain this pattern in stderr.
            if 'api4 [--in IN] [--out OUT]'.lower() in exc.stderr.lower():
                raise InvalidApiCall(exc)

            # For everything else we UnkownConsoleError.
            else:
                raise exc

        else:
            logger.info(f'Command result: {reply}')
            return reply


class ConsoleApiV3(BaseConsoleApi):
    """
    Console API bindings for CiviCRM APIv3.

    :param str cv: cv command
    :param str cwd: working directory for cv
    :param str context:
        If a context is given the original command will be tokenized and
        given to the context as positional argument. The simplest context
        migth be a 'sh -c'. But it could also be a docker-exec- or
        ssh-command to run the api call within a docker container or on a
        remote machine.
    """
    VERSION = v3

    def _get_command(self, entity, action, params):
        params['sequential'] = 1
        echo_params = ['echo', shlex.quote(json.dumps(params))]
        api = ['api3', shlex.quote(f'{entity}.{action}'), '--in=json']
        command = self.cv + self.cwd + api
        return '{} | {}'.format(' '.join(echo_params), ' '.join(command))


class ConsoleApiV4(BaseConsoleApi):
    """
    Console API bindings for CiviCRM APIv4.

    :param str cv: cv command
    :param str cwd: working directory for cv
    :param str context:
        If a context is given the original command will be tokenized and
        given to the context as positional argument. The simplest context
        migth be a 'sh -c'. But it could also be a docker-exec- or
        ssh-command to run the api call within a docker container or on a
        remote machine.
    """
    VERSION = v4

    def _get_command(self, entity, action, params):
        api = ['api4', shlex.quote(f'{entity}.{action}')]
        params = [shlex.quote(json.dumps(params))] if params else list()
        command = self.cv + self.cwd + api + params
        return ' '.join(command)




"""
This module defines default entities and actions for CiviCRM APIv4.
"""

ENTITIES = [
    'ACL',
    'ACLEntityRole',
    'ActionSchedule',
    'Activity',
    'ActivityContact',
    'Address',
    'Batch',
    'Contact',
    'ContactType',
    'Contribution',
    'ContributionPage',
    'ContributionProduct',
    'ContributionRecur',
    'ContributionSoft',
    'Country',
    'County',
    'CustomField',
    'CustomGroup',
    'Dashboard',
    'DashboardContact',
    'DedupeException',
    'DedupeRule',
    'DedupeRuleGroup',
    'Domain',
    'Email',
    'Entity',
    'EntityBatch',
    'EntityFinancialAccount',
    'EntityFinancialTrxn',
    'EntityTag',
    'Event',
    'ExampleData',
    'Extension',
    'File',
    'FinancialAccount',
    'FinancialItem',
    'FinancialTrxn',
    'FinancialType',
    'Group',
    'GroupContact',
    'GroupNesting',
    'GroupOrganization',
    'IM',
    'Job',
    'JobLog',
    'LineItem',
    'LocBlock',
    'LocationType',
    'Log',
    'MailSettings',
    'Mailing',
    'MailingEventBounce',
    'MailingEventConfirm',
    'MailingEventOpened',
    'MailingEventQueue',
    'MailingEventSubscribe',
    'MailingEventUnsubscribe',
    'MailingGroup',
    'MailingJob',
    'Managed',
    'Mapping',
    'MappingField',
    'Membership',
    'MembershipBlock',
    'MembershipStatus',
    'MembershipType',
    'MessageTemplate',
    'Navigation',
    'Note',
    'OpenID',
    'OptionGroup',
    'OptionValue',
    'PCP',
    'PCPBlock',
    'Participant',
    'PaymentProcessor',
    'PaymentProcessorType',
    'PaymentToken',
    'Permission',
    'Phone',
    'PriceField',
    'PriceFieldValue',
    'PriceSet',
    'PriceSetEntity',
    'PrintLabel',
    'Product',
    'Queue',
    'RecentItem',
    'Relationship',
    'RelationshipCache',
    'RelationshipType',
    'ReportInstance',
    'Route',
    'SavedSearch',
    'SearchDisplay',
    'SearchSegment',
    'Setting',
    'StateProvince',
    'StatusPreference',
    'SubscriptionHistory',
    'System',
    'Tag',
    'Translation',
    'UFField',
    'UFGroup',
    'UFJoin',
    'UFMatch',
    'UserJob',
    'Website',
    'WordReplacement',
    'WorkflowMessage',
    'WorldRegion',
]
"""
All standard entities for CiviCRM APIv4.
"""

ACTIONS = [
    'checkAccess',
    'getActions',
    'getFields',
    'get',
    'create',
    'update',
    'save',
    'delete',
    'replace',
]
"""
All standard actions for CiviCRM APIv4.
"""



import logging
import json
from . import v4
from .utils import dict_to_where_clause_list
from .errors import InvalidApiCall
from .errors import InvalidResponse


logger = logging.getLogger('civicrmapi')


class BaseAction:
    """
    Base class for CiviCRM API actions. Subclasses must define the
    :attr:`~BaseAction.ENTITY` and :attr:`~BaseAction.ACTION` attributes.

    By calling an instance of this class the specific api call will be
    performed. See :meth:`~BaseAction.__call__`.

    :param api: CiviCRM API instance
    :type api: :class:`~civicrmapi.base.BaseApi`
    :raises NotImplemented: when :attr:`~BaseAction.ENTITY` is not defined
    :raises NotImplemented: when :attr:`~BaseAction.ACTION` is not defined
    """
    ENTITY = None
    """
    The CiviCRM entity this action operates on. Set by a subclass of
    :class:`~.BaseAction`.
    """
    ACTION = None
    """
    The action name. Set by a subclass of :class:`~.BaseAction`.
    """

    def __init__(self, api):
        self._api = api

        # Raise NotImplemented if ENTITY or ACTIONis not defined.
        if not self.ENTITY or not self.ACTION:
            raise NotImplemented('ENTITY and ACTION must be defined.')

    def __call__(self, params=None):
        """
        Perform the api call.

        :param dict params: api call parameters (optional)
        :return dict: api call result
        """
        return self._api(self.ENTITY, self.ACTION, params or dict())


class BaseEntity:
    """
    Base class for CiviCRM entities. Subclasses must define the
    :attr:`~BaseEntity.ENTITY` attribute.

    This class will be initialized with all default actions defined for the
    specific api version as attributes.

    By calling an instance of this class the specific api call will be
    performed. See :meth:`~BaseEntity.__call__`.

    :param api: CiviCRM API instance
    :type api: :class:`~civicrmapi.base.BaseApi`
    :raises NotImplemented: when :attr:`~BaseEntity.ENTITY` is not defined
    """

    ENTITY = None
    """
    The entity name. Set by a subclass of :class:`~.BaseEntity`.
    """

    def __init__(self, api):
        self._api = api

        # Raise NotImplemented if ENTITY is not defined.
        if not self.ENTITY:
            raise NotImplemented('ENTITY must be defined.')

        # Add default actions defined in the api version module.
        for action_name in self._api.VERSION.ACTIONS:
            attrs = dict(ENTITY=self.ENTITY, ACTION=action_name)
            action_class = type(action_name, (BaseAction,), attrs)
            setattr(self, action_name, action_class(self._api))

    def __call__(self, action, params=None):
        """
        Perform an api call on this entity.

        :param str action: api call action
        :param dict params: api call parameters (optional)
        :return dict: api call result
        """
        return self._api(self.ENTITY, action, params or dict())


class BaseApi:
    """
    Base CiviCRM API class. Subclasses must define the :attr:`~BaseApi.VERSION`
    attribute and overwrite the :meth:`~BaseApi._perform_api_call` method.

    This class will be initialized with the standard entities for the specific
    api version as instance attributes. (See :mod:`~civicrmapi.v3` and
    :mod:`~civicrmapi.v4`.)

    By calling an instance of this class the specific api call will be
    performed. See :meth:`~BaseApi.__call__`.

    :raises NotImplemented: when :attr:`~BaseApi.VERSION` is not defined
    :raises NotImplemented: when :meth:`._perform_api_call` is not implemented
    """

    VERSION = None
    """
    The CiviCRM API version. Use either :mod:`~civicrmapi.v3` or
    :mod:`~civicrmapi.v4` in your subclass.
    """

    # FIXME: Add standard entities in __new__(), while extra entities in __init__()?
    def __init__(self):
        # Raise NotImplemented if VERSION is not defined.
        if not self.VERSION:
            raise NotImplemented('VERSION must be defined.')

        # Add default entities defined in the api version module.
        for entity in self.VERSION.ENTITIES:
            entity_class = type(entity, (BaseEntity,), dict(ENTITY=entity))
            setattr(self, entity, entity_class(self))

    def __call__(self, entity, action, params=None):
        """
        Perform an api call.

        :param str entity: CiviCRM-entitiy
        :param str action: api call action
        :param dict params: api call parameters (optional)
        :return dict: normalized api call result
        """
        logger.info(f'Perform api call: {entity}.{action} with {params}')
        params = self._prepare_api_v4_parameters(action, params)
        response = self._perform_api_call(entity, action, params or dict())
        data = self._get_data_from_response(response)
        data = self._check_api_response(data)
        return self._normalize_result_values(data)

    def _prepare_api_v4_parameters(self, action, params):
        """
        The `API v3`_ and `API v4`_ have different formats for their api
        parameters. While the API v3 uses a flat dictonary with entity field
        parameters as well as meta informations like `limit` on the same level,
        the API v4 has a more structured and "sql-like" way to format the
        parameters - using keywords like `values`, `where` and `join`.

        .. _API v3: https://docs.civicrm.org/dev/en/latest/api/v3/usage/
        .. _API v4: https://docs.civicrm.org/dev/en/latest/api/v4/usage/

        For complex api calls you will need to lookup the api version's specific
        way on how to build your parameters dict. (Use the api explorer of a
        civicrm installation or just read the `docs`_.)

        .. _docs: https://docs.civicrm.org/dev/

        Simple api calls using only entity field parameters to fetch or create
        objects can be passed in as they are. This method will ensure that
        those parameters are prepared for the API v4 adding them to a `values`
        or `where` key.

        :param dict params: api parameters
        :return dict: prepared api parameters
        """
        if params and self.VERSION is v4:
            if action in ['get', 'delete'] and not 'where' in params:
                params = dict(where=dict_to_where_clause_list(params))
            elif action == 'create' and not 'values' in params:
                params = dict(values=params)
            elif action == 'update' and not 'where' in params and 'id' in params:
                id = params.pop('id')
                params = dict(where=dict_to_where_clause_list(dict(id=id)), values=params)
        return params

    def _perform_api_call(self, entity, action, params):
        """
        Must be implemented by subclasses.

        :param str entity: CiviCRM-entitiy
        :param str action: api call action
        :param dict params: api call parameters (optional)
        :return dict: api response data
        """
        raise NotImplemented

    def _get_data_from_response(self, response):
        """
        Takes a response object (either a requests response instance or a
        subprocess completed process instance) and try to parse it as json data.

        :param response: response instance
        :type response: requests.Response or subprocess.CompletedProcess
        :return dict: api response data

        :raises InvalidResponse: when the response value is no valid json
        """
        value = response.text if hasattr(response, 'text') else response.stdout
        try:
            data = json.loads(value)
        except json.JSONDecodeError:
            raise InvalidResponse(response)
        else:
            return data

    def _check_api_response(self, data):
        """
        Process the json response from an api call.

        :param dict data: api response data
        :return dict: api response data

        :raises InvalidApiCall: when an error code is found in the api response
        """
        # Check api-v3 data for an api error.
        if isinstance(data, dict) and data.get('is_error', False):
            raise InvalidApiCall(data)
        # Check api-v4 data for an api error.
        elif isinstance(data, dict) and 'error_code' in data:
            raise InvalidApiCall(data)
        else:
            return data

    def _normalize_result_values(self, data):
        """
        Normalize api v3 and v4 response. Always return api results as a list.

        :param dict data: api response data
        :return dict: normalized api response data

        :raises InvalidResponse: when data has not the expected type or keys.
        """
        # Console api v4 returns values as list.
        if isinstance(data, list):
            return data
        # If we have a dict with values we return the values list instead.
        elif 'values' in data:
            return data['values']
        # Otherwise something is weird. We raise InvalidResponse exception.
        else:
            raise InvalidResponse(data)



"""
----------
civicrmapi
----------
This package provides a convenient way to access `CiviCRM's API`_ from Python
applications. It supports both API verions `v3`_ and `v4`_ as REST API and via
the console command `cv`_.

.. _CiviCRM's API: https://docs.civicrm.org/dev/en/latest/api/
.. _v3: https://docs.civicrm.org/dev/en/latest/api/v3/usage/
.. _v4: https://docs.civicrm.org/dev/en/latest/api/v4/usage/
.. _cv: https://github.com/civicrm/cv


Getting Started
---------------
There are four ready to use api classes:

- :class:`~civicrmapi.rest.RestApiV3` - REST API bindings for CiviCRM APIv3
- :class:`~civicrmapi.rest.RestApiV4` - REST API bindings for CiviCRM APIv4
- :class:`~civicrmapi.console.ConsoleApiV3` - Using `cv` to access CiviCRM APIv3
- :class:`~civicrmapi.console.ConsoleApiV4` - Using `cv` to access CiviCRM APIv4

All you need to do is to initialize the api of your choice and use it::

    from civicrmapi import ConsoleApiV4

    api = ConsoleApiV4(cv='/path/to/cv', cwd='/path/to/civicrm/root')
    params = {
        "contact_type": "Organization",
        "organization_name": "pretty org",
        }
    result = api.Contact.create(params)


Concept
-------
There are three main base classes an api binding is built of:

- :class:`~base.BaseApi` - Base class for all API implementations
- :class:`~base.BaseEntity` - Base class for CiviCRM entities
- :class:`~base.BaseAction` - Base class for CiviCRM actions

Initializing an api class will automatically equip it with entities, which
themselves are equipped with actions - each as instance attributes. And on each
level of this hierarchy an api call can be performed. The api, its entities and
their actions are all callable.

The following ways of performing an api call are equivalent::

    from civicrmapi import RestApiV4

    api = RestApiV4(url='https://example.org/civicrm', api_key='your_api_key')
    params = {"contact_type": "Organization"}

    # Calling the action instance.
    result = api.Contact.get(params)

    # Calling the entity instance passing the action as parameter.
    result = api.Contact('get', params)

    # Calling the api instance  passing the entity and action as parameters.
    result = api('Contact', 'get', params)


The standard entities and actions for each api version are defined within the
the corresponding modules :mod:`~.v3` and :mod:`~.v4`. You can also implement
special behavior like custom validation, result parsing or error handling by
writing your own api, entity or action subclasses.


API Parameters
--------------
The `API v3`_ and `API v4`_ have different formats for api parameters. While the
API v3 uses a flat dictonary with entity field parameters as well as meta
informations like `limit` on the same level, the API v4 has a more structured
and "sql-like" way to format the parameters - using keywords like `select`,
`values`, `where` and `join`.

.. _API v3: https://docs.civicrm.org/dev/en/latest/api/v3/usage/
.. _API v4: https://docs.civicrm.org/dev/en/latest/api/v4/usage/

You can always use the version's specific way of structuring the parameters for
your api call. Just use the `api explorer`_ or read the `api documentation`_ to
check out what's possible and how to achieve it.

.. _api explorer: https://docs.civicrm.org/dev/en/latest/api/#api-explorer
.. _api documentation: https://docs.civicrm.org/dev/en/latest/api/

For simple api calls using only entity field parameters to get, create, delete
or update objects, civicrmapi helps you building the v4 parameters out of a
plain field-value dictonary.

Let's say you want to get (or delete) all Individuals with german as their
preferred language. The original v4 parameters would be::

    params = {
        'where': [
            ['contact_type', '=', 'Individual'],
            ['preferred_language', '=', 'de_DE'],
        ]
    }


Those parameters can also be simply written as::

    params = {
        'contact_type': 'Individual',
        'preferred_language: 'de_DE',
    }


The same works for a create api call::

    # Original v4 parameters:
    params = {
        'values': {
            'contact_type': 'Organization',
            'organization_name': 'Super Org',
        }
    }

    # Could be also passed in as:
    params = {
        'contact_type': 'Organization',
        'organization_name: 'Super Org',
    }


And even for an update api call if you use the id field to select your entity::

    # Original v4 parameters:
    params = {
        'where': [
            ['id', '=', 123],
        ],
        'values': {
            'organization_name': 'Mega Org',
        }
    }

    # Could be also passed in as:
    params = {
        'id': 123,
        'organization_name: 'Mega Org',
    }


Results
-------
The REST-API-v3 and v4 as well as the CV-API-v3 returns a dictonary with some
meta data and a values key which holds the result data. Only the CV-API-v4
returns the results as list.

Since the meta data of the result dictonaries are reduntant (such as api
version, action being used, count of the results and other stuff) civicrmapi
skip them and returns for all APIs a plain list of results. So you always get
something like::

    [{'id': 1, ...}, {'id': 2, ...}, ...]


"""

from .base import BaseAction
from .base import BaseEntity
from .base import BaseApi
from .errors import InvalidApiCall
from .errors import InvalidResponse
from .rest import RestApiV3
from .rest import RestApiV4
from .console import ConsoleApiV3
from .console import ConsoleApiV4


"""
Exceptions raised by civicrmapi.

There are three main levels where errors can occur when performing an api call:

- The api cannot be reached.
- Invalid api credentials are used.
- The api call is invalid due to malformed parameters.
- The api call cannot be processed by CiviCRM..

Exceptions raised by the first level are subprocess or requests related. We
leave them alone. If we got valid Responses but do not get valid json formatted
result data we raise InvalidResponse. This for sure can have many different
reasons.

For all api related exceptions (level 2, 3 and 4) we use the InvalidApiCall
class. (We try hard to distinguish them from not api related ones... which is
not always easy.)
"""

import requests
import subprocess
from .utils import format_requests_response
from .utils import format_subprocess_response


class InvalidApiCall(Exception):
    """
    Raised for invalid api calls of any kind:

    - Access denial because of invalid credentials.
    - Malformed api calls.
    - Invalid parameters.
    - All kind of errors due to CiviCRM's buisness logic and database integrity.
    """
    def __init__(self, value):
        if isinstance(value, dict):
            value = value.get('error_message', str(value))
        elif isinstance(value, subprocess.CalledProcessError):
            value = value.stderr
        super().__init__(value)


class InvalidResponse(Exception):
    """
    Raised for valid responses with for - which reason ever - invalid response
    data. For example a subprocess call which succeeds but do not return valid
    json data. This may happen if the command does not reach the api at all, but
    do not exit with a proper return code.

    :params value: Either requests or subprocess response or a data dict
    :type value: dict or requests.Response or subprocess.CompletedProcess
    """
    def __init__(self, value):
        msg = 'Invalid result data:\n'
        if isinstance(value, dict):
            msg += str(value)
        elif isinstance(value, requests.Response):
            msg += format_requests_response(value)
        elif isinstance(value, subprocess.CompletedProcess):
            msg += format_subprocess_response(value)
        super().__init__(msg)




def dict_to_where_clause_list(data):
    where_params = list()
    for key, value in data.items():
        where_params.append([key, '=', value])
    return where_params


def format_requests_response(value):
    msg = f'URL: {value.url}\n'
    msg += f'STATUS: {value.status_code}\n'
    msg += f'CONTENT: {value.text}\n'
    return msg


def format_subprocess_response(value):
    msg = f'RETURN CODE: {value.returncode}\n'
    if value.stdout:
        msg += f'STDOUT: {value.stdout}\n'
    if value.stderr:
        msg += f'STDERR: {value.stderr}\n'
    return msg
